# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest

from lsst.daf.relation import Transfer, iteration, tests


class TransferTestCase(tests.RelationTestCase):
    """Tests for the Transfer operation and relations based on it."""

    def setUp(self) -> None:
        self.a = tests.ColumnTag("a")
        self.source_engine = iteration.Engine(name="source")
        self.destination_engine = iteration.Engine(name="destination")
        self.leaf = self.source_engine.make_leaf(
            {self.a}, payload=iteration.RowSequence([{self.a: 0}, {self.a: 1}]), name="leaf"
        )

    def test_attributes(self) -> None:
        """Check that all UnaryOperation and Relation attributes have the
        expected values.
        """
        relation = self.leaf.transferred_to(self.destination_engine)
        assert isinstance(relation, Transfer)
        self.assertEqual(relation.columns, self.leaf.columns)
        self.assertEqual(relation.engine, self.destination_engine)
        self.assertEqual(relation.min_rows, self.leaf.min_rows)
        self.assertEqual(relation.max_rows, self.leaf.max_rows)
        self.assertFalse(relation.is_locked)
        self.assertTrue(relation.destination, self.destination_engine)

    def test_apply_simplify(self) -> None:
        """Test simplification logic in Transfer.apply."""
        self.assert_relations_equal(self.leaf.transferred_to(self.source_engine), self.leaf)
        self.assert_relations_equal(
            self.leaf.transferred_to(self.destination_engine).transferred_to(self.source_engine), self.leaf
        )

    def test_str(self) -> None:
        """Test str(Transfer) and
        str(UnaryOperationRelation[Transfer]).
        """
        relation = self.leaf.transferred_to(self.destination_engine)
        self.assertEqual(str(relation), f"→[destination]({self.leaf})")


if __name__ == "__main__":
    unittest.main()



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest

from lsst.daf.relation import LeafRelation, iteration, tests


class LeafRelationTestCase(unittest.TestCase):
    """Tests for LeafRelation and closely-related iteration-engine methods."""

    def setUp(self) -> None:
        self.a = tests.ColumnTag("a")
        self.b = tests.ColumnTag("b")

    def test_iteration_make_leaf(self) -> None:
        """Test the iteration engine's `make_leaf` function, and with it
        the dataclass field definitions for `LeafRelation` itself`.
        """
        engine = iteration.Engine()
        columns = {self.a, self.b}
        sequence_payload = iteration.RowSequence(
            [{self.a: 0, self.b: 0}, {self.a: 0, self.b: 1}, {self.a: 1, self.b: 0}, {self.a: 0, self.b: 0}]
        )
        mapping_payload = sequence_payload.to_mapping((self.a, self.b))
        sequence_leaf = engine.make_leaf(columns, payload=sequence_payload)
        self.assertEqual(sequence_leaf.engine, engine)
        self.assertEqual(sequence_leaf.columns, columns)
        self.assertEqual(sequence_leaf.min_rows, 4)
        self.assertEqual(sequence_leaf.max_rows, 4)
        self.assertTrue(sequence_leaf.is_locked)
        self.assertFalse(sequence_leaf.is_join_identity)
        self.assertFalse(sequence_leaf.is_trivial)
        self.assertIsNone(sequence_leaf.parameters)
        self.assertTrue(sequence_leaf.name.startswith("leaf"))
        self.assertCountEqual(sequence_leaf.payload, sequence_payload)
        mapping_leaf = engine.make_leaf(columns=columns, payload=mapping_payload)
        self.assertEqual(mapping_leaf.engine, engine)
        self.assertEqual(mapping_leaf.columns, columns)
        self.assertEqual(mapping_leaf.min_rows, 3)
        self.assertEqual(mapping_leaf.max_rows, 3)
        self.assertTrue(mapping_leaf.is_locked)
        self.assertFalse(mapping_leaf.is_join_identity)
        self.assertFalse(mapping_leaf.is_trivial)
        self.assertIsNone(mapping_leaf.parameters)
        self.assertTrue(mapping_leaf.name.startswith("leaf"))
        self.assertCountEqual(mapping_leaf.payload, mapping_payload)
        self.assertNotEqual(mapping_leaf.name, sequence_leaf.name)
        self.assertNotEqual(mapping_leaf, sequence_leaf)

    def test_join_identity(self) -> None:
        """Test `LeafRelation.make_join_identity and the iteration engine's
        get_join_identity_payload."""
        engine = iteration.Engine()
        join_identity_leaf = LeafRelation.make_join_identity(engine)
        self.assertEqual(join_identity_leaf, LeafRelation.make_join_identity(engine))
        self.assertEqual(join_identity_leaf.engine, engine)
        self.assertEqual(join_identity_leaf.columns, set())
        self.assertEqual(join_identity_leaf.min_rows, 1)
        self.assertEqual(join_identity_leaf.max_rows, 1)
        self.assertTrue(join_identity_leaf.is_locked)
        self.assertTrue(join_identity_leaf.is_join_identity)
        self.assertTrue(join_identity_leaf.is_trivial)
        self.assertIsNone(join_identity_leaf.parameters)
        self.assertEqual(join_identity_leaf.name, "I")
        self.assertCountEqual(join_identity_leaf.payload, [{}])

    def test_doomed(self) -> None:
        """Test `LeafRelation.make_doomed and the iteration engine's
        get_doomed_payload."""
        engine = iteration.Engine()
        doomed_leaf = LeafRelation.make_doomed(engine, {self.a, self.b}, ["doomed 1"])
        self.assertEqual(
            doomed_leaf,
            LeafRelation(
                engine,
                frozenset({self.a, self.b}),
                [],
                messages=["doomed 1"],
                name="0",
                min_rows=0,
                max_rows=0,
            ),
        )
        # Messages not important for equality.
        self.assertEqual(doomed_leaf, LeafRelation.make_doomed(engine, {self.a, self.b}, ["doomed 2"]))
        # Columns are important for equality.
        self.assertNotEqual(doomed_leaf, LeafRelation.make_doomed(engine, {self.a}, ["doomed 1"]))
        # Name is important for equality.
        self.assertNotEqual(
            doomed_leaf, LeafRelation.make_doomed(engine, {self.a, self.b}, ["doomed 1"], name="c")
        )
        self.assertEqual(doomed_leaf.engine, engine)
        self.assertEqual(doomed_leaf.columns, {self.a, self.b})
        self.assertEqual(doomed_leaf.min_rows, 0)
        self.assertEqual(doomed_leaf.max_rows, 0)
        self.assertTrue(doomed_leaf.is_locked)
        self.assertFalse(doomed_leaf.is_join_identity)
        self.assertTrue(doomed_leaf.is_trivial)
        self.assertIsNone(doomed_leaf.parameters, None)
        self.assertEqual(doomed_leaf.name, "0")
        self.assertCountEqual(doomed_leaf.payload, [])

    def test_bad_min_max_rows(self) -> None:
        """Test construction checks for inconsistent min_rows / max_rows."""
        engine = iteration.Engine()
        with self.assertRaises(ValueError):
            LeafRelation(engine, frozenset({self.a, self.b}), payload=..., min_rows=2, max_rows=1)

    def test_str(self) -> None:
        """Test str(LeafRelation)."""
        engine = iteration.Engine()
        self.assertEqual(
            str(LeafRelation(engine, frozenset({self.a, self.b}), payload=..., name="leaf1")), "leaf1"
        )
        self.assertEqual(
            str(
                LeafRelation(
                    engine, frozenset({self.a, self.b}), payload=..., name="leaf2", parameters=[1, 2]
                )
            ),
            "leaf2([1, 2])",
        )


if __name__ == "__main__":
    unittest.main()



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest
from typing import TypeAlias

import sqlalchemy
from lsst.daf.relation import (
    ColumnExpression,
    ColumnTag,
    Identity,
    IgnoreOne,
    Relation,
    RelationalAlgebraError,
    SortTerm,
    sql,
    tests,
)

_L: TypeAlias = sqlalchemy.sql.ColumnElement


def _make_leaf(
    engine: sql.Engine, md: sqlalchemy.schema.MetaData, name: str, *columns: ColumnTag
) -> Relation:
    """Make a leaf relation in a SQL engine, backed by SQLAlchemy tables.

    This doesn't actually create tables in any database (even in memory); it
    just creates their SQLAlchemy representations so their names and columns
    render as expected in SQL strings.

    Parameters
    ----------
    engine : `lsst.daf.relation.sql.Engine`
        Relation engine for the new leaf.
    md : `sqlalchemy.schema.MetaData`
        SQLAlchemy metadata object to add tables to.
    name : `str`
        Name of the relation and its table.
    *columns: `ColumnTag`
        Columns to include in the relation and its table.

    Returns
    -------
    leaf : `Relation`
        Leaf relation backed by the new table.
    """
    columns_available = {
        tag: sqlalchemy.schema.Column(tag.qualified_name, sqlalchemy.Integer) for tag in columns
    }
    table = sqlalchemy.schema.Table(name, md, *columns_available.values())
    payload = sql.Payload[_L](from_clause=table, columns_available=columns_available)
    return engine.make_leaf(columns_available.keys(), payload=payload, name=name)


class SqlEngineTestCase(tests.RelationTestCase):
    def setUp(self):
        self.maxDiff = None

    def test_select_operations(self) -> None:
        """Test SQL engine conversion for different combinations and
        permutations of the operation types managed by the `Select` marker
        relation.
        """
        engine = sql.Engine[_L]()
        md = sqlalchemy.schema.MetaData()
        a = tests.ColumnTag("a")
        b = tests.ColumnTag("b")
        c = tests.ColumnTag("c")
        d = tests.ColumnTag("d")
        expression = ColumnExpression.reference(b).method("__neg__")
        predicate = ColumnExpression.reference(c).gt(ColumnExpression.literal(0))
        terms = [SortTerm(ColumnExpression.reference(b))]
        leaf1 = _make_leaf(engine, md, "leaf1", a, b)
        leaf2 = _make_leaf(engine, md, "leaf2", a, c)
        r = leaf1.with_calculated_column(d, expression).join(leaf2.with_rows_satisfying(predicate))
        self.check_sql_str(
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0",
            engine.to_executable(r),
        )
        # Add modifiers to that query via relation operations.
        self.check_sql_str(
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0",
            engine.to_executable(r.without_duplicates()),
        )
        self.check_sql_str(
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0",
            engine.to_executable(r.with_only_columns({b, c, d})),
        )
        self.check_sql_str(
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b",
            engine.to_executable(r.sorted(terms)),
        )
        self.check_sql_str(
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 3 OFFSET 2",
            engine.to_executable(r[2:5]),
        )
        # Add both a Projection and then a Deduplication.
        self.check_sql_str(
            "SELECT DISTINCT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0",
            engine.to_executable(r.with_only_columns({b, c, d}).without_duplicates()),
        )
        # Add a Deduplication and then a Projection, which requires a subquery.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0"
            ") AS anon_1",
            engine.to_executable(r.without_duplicates().with_only_columns({b, c, d})),
        )
        # Projection and Sort together, in any order.
        self.check_sql_str(
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b",
            engine.to_executable(r.with_only_columns({b, c, d}).sorted(terms)),
            engine.to_executable(r.sorted(terms).with_only_columns({b, c, d})),
        )
        # Projection and Slice together, in any order.
        self.check_sql_str(
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1",
            engine.to_executable(r.with_only_columns({b, c, d})[1:3]),
            engine.to_executable(r[1:3].with_only_columns({b, c, d})),
        )
        # Deduplication and Sort together, in any order.
        self.check_sql_str(
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b",
            engine.to_executable(r.without_duplicates().sorted(terms)),
            engine.to_executable(r.sorted(terms).without_duplicates()),
        )
        # Deduplication and then Slice.
        self.check_sql_str(
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1",
            engine.to_executable(r.without_duplicates()[1:3]),
        )
        # Slice and then Deduplication, which requires a subquery.
        self.check_sql_str(
            "SELECT DISTINCT anon_1.a AS a, anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1",
            engine.to_executable(r[1:3].without_duplicates()),
        )
        # Sort and then Slice.
        self.check_sql_str(
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b LIMIT 2 OFFSET 1",
            engine.to_executable(r.sorted(terms)[1:3]),
        )
        # Slice and then Sort, which requires a subquery.
        self.check_sql_str(
            "SELECT anon_1.a AS a, anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r[1:3].sorted(terms)),
        )
        # Projection then Deduplication, with Sort at any point since it should
        # commute with both.
        self.check_sql_str(
            "SELECT DISTINCT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b",
            engine.to_executable(r.with_only_columns({b, c, d}).sorted(terms).without_duplicates()),
            engine.to_executable(r.with_only_columns({b, c, d}).without_duplicates().sorted(terms)),
            engine.to_executable(r.sorted(terms).with_only_columns({b, c, d}).without_duplicates()),
        )
        # Deduplication then Projection (via a subquery), with a Sort at any
        # point - but the engine will need to make sure the Sort always appears
        # in the outer query, since the subquery does not preserve order.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r.without_duplicates().with_only_columns({b, c, d}).sorted(terms)),
            engine.to_executable(r.without_duplicates().sorted(terms).with_only_columns({b, c, d})),
            engine.to_executable(r.sorted(terms).without_duplicates().with_only_columns({b, c, d})),
        )
        # Projection then Deduplication then Slice.
        self.check_sql_str(
            "SELECT DISTINCT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1",
            engine.to_executable(r.with_only_columns({b, c, d}).without_duplicates()[1:3]),
        )
        # Projection and Slice in any order, then Deduplication, which requires
        # a subquery.
        self.check_sql_str(
            "SELECT DISTINCT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1",
            engine.to_executable(r.with_only_columns({b, c, d})[1:3].without_duplicates()),
            engine.to_executable(r[1:3].with_only_columns({b, c, d}).without_duplicates()),
        )
        # Deduplication then Projection and Slice in any order, which requires
        # a subquery.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0"
            ") AS anon_1 LIMIT 2 OFFSET 1",
            engine.to_executable(r.without_duplicates().with_only_columns({b, c, d})[1:3]),
            engine.to_executable(r.without_duplicates()[1:3].with_only_columns({b, c, d})),
        )
        # Slice then Deduplication then Projection, which requires two
        # subqueries.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT DISTINCT anon_2.a AS a, anon_2.b AS b, anon_2.c AS c, anon_2.d AS d FROM ("
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_2"
            ") AS anon_1",
            engine.to_executable(r[1:3].without_duplicates().with_only_columns({b, c, d})),
        )
        # Sort then Slice, with Projection anywhere.
        self.check_sql_str(
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b LIMIT 2 OFFSET 1",
            engine.to_executable(r.sorted(terms)[1:3].with_only_columns({b, c, d})),
            engine.to_executable(r.sorted(terms).with_only_columns({b, c, d})[1:3]),
            engine.to_executable(r.with_only_columns({b, c, d}).sorted(terms)[1:3]),
        )
        # Slice then Sort then Projection, which requires a subquery.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r[1:3].sorted(terms).with_only_columns({b, c, d})),
        )
        # Slice and Projection in any order, then Sort, which requires a
        # subquery.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r[1:3].with_only_columns({b, c, d}).sorted(terms)),
            engine.to_executable(r.with_only_columns({b, c, d})[1:3].sorted(terms)),
        )
        # Deduplication and Sort in any order, then Slice.
        self.check_sql_str(
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b "
            "LIMIT 2 OFFSET 1",
            engine.to_executable(r.without_duplicates().sorted(terms)[1:3]),
            engine.to_executable(r.sorted(terms).without_duplicates()[1:3]),
        )
        # Duplication then Slice then Sort, which requires a subquery.
        self.check_sql_str(
            "SELECT anon_1.a AS a, anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r.without_duplicates()[1:3].sorted(terms)),
        )
        # Slice then Sort and Deduplication in any order, which requires a
        # subquery.
        self.check_sql_str(
            "SELECT DISTINCT anon_1.a AS a, anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r[1:3].without_duplicates().sorted(terms)),
            engine.to_executable(r[1:3].sorted(terms).without_duplicates()),
        )
        # Projection then Deduplication, with Sort at any point, and finally a
        # Slice.
        self.check_sql_str(
            "SELECT DISTINCT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b LIMIT 2 OFFSET 1",
            engine.to_executable(r.with_only_columns({b, c, d}).sorted(terms).without_duplicates()[1:3]),
            engine.to_executable(r.with_only_columns({b, c, d}).without_duplicates().sorted(terms)[1:3]),
            engine.to_executable(r.sorted(terms).with_only_columns({b, c, d}).without_duplicates()[1:3]),
        )
        # Projection, Deduplication, Slice, Sort.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT DISTINCT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r.with_only_columns({b, c, d}).without_duplicates()[1:3].sorted(terms)),
        )
        # Projection and Slice in any order, then Deduplication and Sort in any
        # order.
        self.check_sql_str(
            "SELECT DISTINCT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r.with_only_columns({b, c, d})[1:3].without_duplicates().sorted(terms)),
            engine.to_executable(r.with_only_columns({b, c, d})[1:3].sorted(terms).without_duplicates()),
            engine.to_executable(r[1:3].with_only_columns({b, c, d}).without_duplicates().sorted(terms)),
            engine.to_executable(r[1:3].with_only_columns({b, c, d}).sorted(terms).without_duplicates()),
        )
        # Projection and Sort in any order, then Slice, then Deduplication.
        self.check_sql_str(
            "SELECT DISTINCT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b "
            "LIMIT 2 OFFSET 1"
            ") AS anon_1",
            engine.to_executable(r.with_only_columns({b, c, d}).sorted(terms)[1:3].without_duplicates()),
            engine.to_executable(r.sorted(terms).with_only_columns({b, c, d})[1:3].without_duplicates()),
        )
        # Deduplication then Projection (via a subquery), with a Sort at any
        # point, and then finally a Slice.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0"
            ") AS anon_1 ORDER BY anon_1.b LIMIT 2 OFFSET 1",
            engine.to_executable(r.without_duplicates().with_only_columns({b, c, d}).sorted(terms)[1:3]),
            engine.to_executable(r.without_duplicates().sorted(terms).with_only_columns({b, c, d})[1:3]),
            engine.to_executable(r.sorted(terms).without_duplicates().with_only_columns({b, c, d})[1:3]),
        )
        # Deduplication, then Projection and Slice in any order, then Sort.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT anon_2.b AS b, anon_2.c AS c, anon_2.d AS d FROM ("
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0"
            ") AS anon_2 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r.without_duplicates().with_only_columns({b, c, d})[1:3].sorted(terms)),
            engine.to_executable(r.without_duplicates()[1:3].with_only_columns({b, c, d}).sorted(terms)),
        )
        # Sort and Deduplication in any order, then Projection and Slice in any
        # order.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT DISTINCT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0"
            ") AS anon_1 ORDER BY anon_1.b LIMIT 2 OFFSET 1",
            engine.to_executable(r.without_duplicates().sorted(terms).with_only_columns({b, c, d})[1:3]),
            engine.to_executable(r.sorted(terms).without_duplicates().with_only_columns({b, c, d})[1:3]),
            engine.to_executable(r.without_duplicates().sorted(terms)[1:3].with_only_columns({b, c, d})),
            engine.to_executable(r.sorted(terms).without_duplicates()[1:3].with_only_columns({b, c, d})),
        )
        # Sort, then Slice and Projection in any order, then Deduplication.
        self.check_sql_str(
            "SELECT DISTINCT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 ORDER BY leaf1.b LIMIT 2 OFFSET 1"
            ") AS anon_1",
            engine.to_executable(r.sorted(terms)[1:3].with_only_columns({b, c, d}).without_duplicates()),
            engine.to_executable(r.sorted(terms).with_only_columns({b, c, d})[1:3].without_duplicates()),
        )
        # Slice, then Projection, then Deduplication and Sort in any order.
        self.check_sql_str(
            "SELECT DISTINCT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r[1:3].with_only_columns({b, c, d}).without_duplicates().sorted(terms)),
            engine.to_executable(r[1:3].with_only_columns({b, c, d}).sorted(terms).without_duplicates()),
        )
        # Slice, then Sort, then Projection, then Deduplication.
        self.check_sql_str(
            "SELECT DISTINCT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r[1:3].sorted(terms).with_only_columns({b, c, d}).without_duplicates()),
        )
        # Slice, then Sort, Deduplication, and Projection as long as the
        # Deduplication precedes the Projection.
        self.check_sql_str(
            "SELECT anon_1.b AS b, anon_1.c AS c, anon_1.d AS d FROM ("
            "SELECT DISTINCT anon_2.a AS a, anon_2.b AS b, anon_2.c AS c, anon_2.d AS d FROM ("
            "SELECT leaf2.a AS a, leaf1.b AS b, leaf2.c AS c, -leaf1.b AS d "
            "FROM leaf1 JOIN leaf2 ON leaf1.a = leaf2.a WHERE leaf2.c > 0 LIMIT 2 OFFSET 1"
            ") AS anon_2"
            ") AS anon_1 ORDER BY anon_1.b",
            engine.to_executable(r[1:3].sorted(terms).without_duplicates().with_only_columns({b, c, d})),
            engine.to_executable(r[1:3].without_duplicates().sorted(terms).with_only_columns({b, c, d})),
            engine.to_executable(r[1:3].without_duplicates().with_only_columns({b, c, d}).sorted(terms)),
        )

    def test_additional_append_unary(self) -> None:
        """Test append_unary rules that involve more than just the
        Select-managed operation types.
        """
        engine = sql.Engine[_L]()
        md = sqlalchemy.schema.MetaData()
        a = tests.ColumnTag("a")
        b = tests.ColumnTag("b")
        c = tests.ColumnTag("c")
        d = tests.ColumnTag("d")
        expression = ColumnExpression.reference(b).method("__neg__")
        predicate = ColumnExpression.reference(c).gt(ColumnExpression.literal(0))
        leaf1 = _make_leaf(engine, md, "leaf1", a, b)
        leaf2 = _make_leaf(engine, md, "leaf2", a, c)
        # Applying a Calculation to a Projection expands the latter and
        # commutes them.
        self.assert_relations_equal(
            leaf1.with_only_columns({b}).with_calculated_column(d, expression),
            leaf1.with_calculated_column(d, expression).with_only_columns({b, d}),
        )
        # Back-to-back Deduplications reduce to one.
        self.assert_relations_equal(
            leaf1.without_duplicates().without_duplicates(),
            leaf1.without_duplicates(),
        )
        # Selections after Slices involve a subquery.
        self.check_sql_str(
            "SELECT anon_1.a AS a, anon_1.c AS c FROM ("
            "SELECT leaf2.a AS a, leaf2.c AS c FROM leaf2 LIMIT 2 OFFSET 1"
            ") AS anon_1 WHERE anon_1.c > 0",
            engine.to_executable(leaf2[1:3].with_rows_satisfying(predicate)),
        )
        # Identity does nothing.
        self.assert_relations_equal(Identity().apply(leaf1), leaf1)

    def test_additional_append_binary(self) -> None:
        """Test append_binary rules that involve more than just the
        Select-managed operation types.
        """
        engine = sql.Engine[_L]()
        md = sqlalchemy.schema.MetaData()
        a = tests.ColumnTag("a")
        b = tests.ColumnTag("b")
        c = tests.ColumnTag("c")
        leaf1 = _make_leaf(engine, md, "leaf1", a, b)
        leaf2 = _make_leaf(engine, md, "leaf2", a, c)
        # Projections are moved outside joins.
        self.assert_relations_equal(
            leaf1.join(leaf2.with_only_columns({a})),
            leaf1.join(leaf2).with_only_columns({a, b}),
        )
        # IgnoreOne does what it should.
        self.assert_relations_equal(IgnoreOne(ignore_lhs=True).apply(leaf1, leaf2), leaf2)
        self.assert_relations_equal(IgnoreOne(ignore_lhs=False).apply(leaf1, leaf2), leaf1)

    def test_chains(self) -> None:
        """Test relation trees that involve Chain operations."""
        engine = sql.Engine[_L]()
        md = sqlalchemy.schema.MetaData()
        a = tests.ColumnTag("a")
        b = tests.ColumnTag("b")
        leaf1 = _make_leaf(engine, md, "leaf1", a, b)
        leaf2 = _make_leaf(engine, md, "leaf2", a, b)
        sort_terms = [SortTerm(ColumnExpression.reference(b))]
        # A Chain on its own maps directly to SQL UNION.
        self.check_sql_str(
            "SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION ALL "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2",
            engine.to_executable(leaf1.chain(leaf2)),
        )
        # Deduplication transforms this to UNION ALL.
        self.check_sql_str(
            "SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2",
            engine.to_executable(leaf1.chain(leaf2).without_duplicates()),
        )
        # Sorting happens after the second SELECT.
        self.check_sql_str(
            "SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION ALL "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2 ORDER BY b",
            engine.to_executable(leaf1.chain(leaf2).sorted(sort_terms)),
        )
        # Slicing also happens after the second SELECT.
        self.check_sql_str(
            "SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION ALL "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2 LIMIT 3 OFFSET 2",
            engine.to_executable(leaf1.chain(leaf2)[2:5]),
        )
        # Projection just before the Chain is handled without any extra
        # subqueries.
        self.check_sql_str(
            "SELECT leaf1.a AS a FROM leaf1 UNION ALL SELECT leaf2.a AS a FROM leaf2",
            engine.to_executable(leaf1.with_only_columns({a}).chain(leaf2.with_only_columns({a}))),
        )
        # Deduplication prior to Chain adds DISTINCT.
        self.check_sql_str(
            "SELECT DISTINCT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION ALL "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2",
            engine.to_executable(leaf1.without_duplicates().chain(leaf2)),
        )
        # Projection and Deduplication prior to Chain also just adds DISTINCT.
        self.check_sql_str(
            "SELECT DISTINCT leaf1.a AS a FROM leaf1 UNION ALL SELECT leaf2.a AS a FROM leaf2",
            engine.to_executable(
                leaf1.with_only_columns({a}).without_duplicates().chain(leaf2.with_only_columns({a}))
            ),
        )
        # Nested chains should flatten out (again, no subqueries), but use
        # parentheses for precedence.
        leaf3 = _make_leaf(engine, md, "leaf3", a, b)
        self.check_sql_str(
            "(SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION ALL "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2) "
            "UNION ALL "
            "SELECT leaf3.a AS a, leaf3.b AS b FROM leaf3",
            engine.to_executable(leaf1.chain(leaf2).chain(leaf3)),
        )
        # Nested chains with deduplication should do the same.
        self.check_sql_str(
            "(SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION ALL "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2) "
            "UNION "
            "SELECT leaf3.a AS a, leaf3.b AS b FROM leaf3",
            engine.to_executable(leaf1.chain(leaf2).chain(leaf3).without_duplicates()),
        )
        self.check_sql_str(
            "(SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2) "
            "UNION "
            "SELECT leaf3.a AS a, leaf3.b AS b FROM leaf3",
            engine.to_executable(leaf1.chain(leaf2).without_duplicates().chain(leaf3).without_duplicates()),
        )
        # Nested chains with projections.
        self.check_sql_str(
            "(SELECT leaf1.a AS a FROM leaf1 "
            "UNION ALL "
            "SELECT leaf2.a AS a FROM leaf2) "
            "UNION ALL "
            "SELECT leaf3.a AS a FROM leaf3",
            engine.to_executable(
                leaf1.chain(leaf2).with_only_columns({a}).chain(leaf3.with_only_columns({a}))
            ),
        )
        # Nested chains with projections and (then) deduplication.
        self.check_sql_str(
            "(SELECT leaf1.a AS a FROM leaf1 "
            "UNION "
            "SELECT leaf2.a AS a FROM leaf2) "
            "UNION "
            "SELECT leaf3.a AS a FROM leaf3",
            engine.to_executable(
                leaf1.chain(leaf2)
                .with_only_columns({a})
                .without_duplicates()
                .chain(leaf3.with_only_columns({a}))
                .without_duplicates()
            ),
        )
        # Chains with Slices and possibly Sorts in operands need subqueries.
        self.check_sql_str(
            "SELECT anon_1.a AS a, anon_1.b AS b FROM "
            "(SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 LIMIT 2 OFFSET 1) AS anon_1 "
            "UNION ALL "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2",
            engine.to_executable(leaf1[1:3].chain(leaf2)),
        )
        self.check_sql_str(
            "SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION ALL "
            "SELECT anon_1.a AS a, anon_1.b AS b FROM "
            "(SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2 LIMIT 2 OFFSET 1) AS anon_1",
            engine.to_executable(leaf1.chain(leaf2[1:3])),
        )
        # Add a Selection or Calculation on top of a Chain yields subqueries
        # to avoid reordering operations.
        expression = ColumnExpression.reference(b).method("__neg__")
        self.check_sql_str(
            "SELECT anon_1.a AS a, anon_1.b AS b, -anon_1.b AS c FROM "
            "(SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION ALL "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2) AS anon_1",
            engine.to_executable(leaf1.chain(leaf2).with_calculated_column(tests.ColumnTag("c"), expression)),
        )
        predicate = ColumnExpression.reference(a).gt(ColumnExpression.literal(0))
        self.check_sql_str(
            "SELECT anon_1.a AS a, anon_1.b AS b FROM "
            "(SELECT leaf1.a AS a, leaf1.b AS b FROM leaf1 "
            "UNION ALL "
            "SELECT leaf2.a AS a, leaf2.b AS b FROM leaf2) AS anon_1 "
            "WHERE anon_1.a > 0",
            engine.to_executable(leaf1.chain(leaf2).with_rows_satisfying(predicate)),
        )

    def test_row_ordering_loss(self) -> None:
        """Test that we raise when we would have to make an existing Sort do
        nothing by putting it in a subquery.
        """
        engine = sql.Engine[_L]()
        md = sqlalchemy.schema.MetaData()
        a = tests.ColumnTag("a")
        b = tests.ColumnTag("b")
        leaf1 = _make_leaf(engine, md, "leaf1", a, b)
        leaf2 = _make_leaf(engine, md, "leaf2", a, b)
        relation = leaf1.sorted([SortTerm(ColumnExpression.reference(b))])
        with self.assertRaises(RelationalAlgebraError):
            relation.materialized()
        with self.assertRaises(RelationalAlgebraError):
            relation.chain(leaf2)
        with self.assertRaises(RelationalAlgebraError):
            leaf2.chain(relation)
        with self.assertRaises(RelationalAlgebraError):
            relation.join(leaf2)
        with self.assertRaises(RelationalAlgebraError):
            leaf2.join(relation)

    def test_trivial(self) -> None:
        """Test that we can emit valid SQL for relations with no columns or
        no rows.
        """
        # No points for pretty; subqueries here are unnecessary but
        # Payload.from_clause would need to be able to be None to drop them,
        # and that's not worth it.
        engine = sql.Engine[_L]()
        join_identity = engine.make_join_identity_relation()
        self.check_sql_str(
            'SELECT 1 AS "IGNORED" FROM (SELECT 1 AS "IGNORED") AS anon_1',
            engine.to_executable(join_identity),
        )
        doomed = engine.make_doomed_relation({tests.ColumnTag("a")}, messages=[])
        self.check_sql_str(
            "SELECT anon_1.a AS a FROM (SELECT NULL AS a) AS anon_1 WHERE 0 = 1",
            engine.to_executable(doomed),
        )


if __name__ == "__main__":
    unittest.main()



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Slice",)

import dataclasses
from collections.abc import Set
from typing import TYPE_CHECKING, Literal, final

from .._columns import ColumnTag
from .._operation_relations import UnaryOperationRelation
from .._unary_operation import Identity, RowFilter, UnaryCommutator, UnaryOperation

if TYPE_CHECKING:
    from .._engine import Engine
    from .._relation import Relation


@final
@dataclasses.dataclass(frozen=True)
class Slice(RowFilter):
    """A relation relation that filters rows that are outside a range of
    positional indices.
    """

    start: int = 0
    """First index to include the output relation (`int`).
    """

    stop: int | None = None
    """One past the last index to include in the output relation
    (`int` or `None`).
    """

    def __post_init__(self) -> None:
        if self.start < 0:
            raise ValueError(f"Slice start {self.start} is negative.")
        if self.stop is not None and self.stop < self.start:
            raise ValueError(f"Slice stop {self.stop} is less than its start {self.start}.")

    @property
    def limit(self) -> int | None:
        """The maximum number of rows to include (`int` or `None`)."""
        return None if self.stop is None else self.stop - self.start

    @property
    def columns_required(self) -> Set[ColumnTag]:
        # Docstring inherited.
        return frozenset()

    @property
    def is_empty_invariant(self) -> Literal[False]:
        # Docstring inherited.
        return False

    @property
    def is_order_dependent(self) -> Literal[True]:
        # Docstring inherited.
        return True

    @property
    def is_count_dependent(self) -> bool:
        # Docstring inherited.
        return True

    def __str__(self) -> str:
        return f"slice[{self.start}:{self.stop}]"

    def _begin_apply(
        self, target: Relation, preferred_engine: Engine | None
    ) -> tuple[UnaryOperation, Engine]:
        # Docstring inherited.
        if not self.start and self.stop is None:
            return Identity(), target.engine
        return super()._begin_apply(target, preferred_engine)

    def _finish_apply(self, target: Relation) -> Relation:
        # Docstring inherited.
        if not self.start and self.stop is None:
            return target
        return super()._finish_apply(target)

    def then(self, next: Slice) -> Slice:
        """Compose this slice with another one.

        Parameters
        ----------
        next : `Slice`
            Slice that acts after ``self``.

        Returns
        -------
        composition : `Slice`
            Slice that is equivalent to ``self`` and ``next`` being applied
            back-to-back.
        """
        new_start = self.start + next.start
        if self.stop is None:
            if next.stop is None:
                new_stop = None
            else:
                new_stop = next.stop + self.start
        else:
            if next.stop is None:
                new_stop = self.stop
            else:
                new_stop = min(self.stop, next.stop + self.start)
        return Slice(new_start, new_stop)

    def applied_min_rows(self, target: Relation) -> int:
        # Docstring inherited.
        if self.stop is not None:
            stop = min(self.stop, target.min_rows)
        else:
            stop = target.min_rows
        return max(stop - self.start, 0)

    def applied_max_rows(self, target: Relation) -> int | None:
        # Docstring inherited.
        if self.stop is not None:
            if target.max_rows is not None:
                stop = min(self.stop, target.max_rows)
            else:
                stop = self.stop
        else:
            if target.max_rows is not None:
                stop = target.max_rows
            else:
                return None
        return max(stop - self.start, 0)

    def commute(self, current: UnaryOperationRelation) -> UnaryCommutator:
        # Docstring inherited.
        from ._calculation import Calculation
        from ._projection import Projection

        match current.operation:
            case Projection() | Calculation():
                return UnaryCommutator(first=self, second=current.operation)
            case _:
                return UnaryCommutator(
                    first=None,
                    second=current.operation,
                    done=False,
                    messages=(
                        f"Slice only commutes with Projection and Calculation, not {current.operation}",
                    ),
                )

    def simplify(self, upstream: UnaryOperation) -> UnaryOperation | None:
        # Docstring inherited.
        if not self.start and self.stop is None:
            return upstream
        match upstream:
            case Slice():
                return upstream.then(self)
        return None



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Join", "PartialJoin")

import dataclasses
from collections.abc import Set
from typing import TYPE_CHECKING, final

from .._binary_operation import BinaryOperation, IgnoreOne
from .._columns import ColumnTag, Predicate
from .._exceptions import ColumnError, EngineError
from .._operation_relations import UnaryOperationRelation
from .._unary_operation import UnaryCommutator, UnaryOperation

if TYPE_CHECKING:
    from .._engine import Engine
    from .._relation import Relation


@final
@dataclasses.dataclass(frozen=True)
class Join(BinaryOperation):
    """A natural join operation.

    A natural join combines two relations by matching rows with the same values
    in their common columns (and satisfying an optional column expression, via
    a `Predicate`), producing a new relation whose columns are the union of the
    columns of its operands.  This is equivalent to [``INNER``] ``JOIN`` in
    SQL.
    """

    predicate: Predicate = dataclasses.field(default_factory=lambda: Predicate.literal(True))
    """A boolean expression that must evaluate to true for any matched rows
    (`Predicate`).

    This does not include the equality constraint on `common_columns`.
    """

    min_columns: frozenset[ColumnTag] = dataclasses.field(default=frozenset())
    """The minimal set of columns that should be used in the equality
    constraint on `common_columns` (`frozenset` [ `ColumnTag` ]).

    If the relations this operation is applied to have common columsn that are
    not a superset of this set, `ColumnError` will be raised by `apply`.

    This is guaranteed to be equal to `max_columns` on any `Join` instance
    attached to a `BinaryOperationRelation` by `apply`.
    """

    max_columns: frozenset[ColumnTag] | None = dataclasses.field(default=None)
    """The maximal set of columns that should be used in the equality
    constraint on `common_columns` (`frozenset` [ `ColumnTag` ] or
    ``None``).

    If the relations this operation is applied to have more columns in common
    than this set, they will not be included in the equality constraint.

    This is guaranteed to be equal to `min_columns` on any `Join` instance
    attached to a `BinaryOperationRelation` by `apply`.
    """

    def __post_init__(self) -> None:
        if self.max_columns is not None and not self.min_columns <= self.max_columns:
            raise ColumnError(
                f"Join min_columns={self.min_columns} is not a subset of max_columns={self.max_columns}."
            )

    @property
    def common_columns(self) -> frozenset[ColumnTag]:
        """The common columns between relations that will be used as an
        equality constraint (`~collections.abc.Set` [ `ColumnTag` ]).

        This attribute is not available on `Join` instances for which
        `min_columns` is not the same as `max_columns`.  It is always available
        on any `Join` instance attached to a `BinaryOperationRelation` by
        `apply`.
        """
        if self.max_columns == self.min_columns:
            return self.min_columns
        else:
            raise ColumnError(f"Common columns for join {self} have not been resolved.")

    def __str__(self) -> str:
        return "⋈"

    def _begin_apply(self, lhs: Relation, rhs: Relation) -> BinaryOperation:
        # Docstring inherited.
        if not self.predicate.columns_required <= self.applied_columns(lhs, rhs):
            raise ColumnError(
                f"Missing columns {set(self.predicate.columns_required - self.applied_columns(lhs, rhs))} "
                f"for join between {lhs!r} and {rhs!r} with predicate {self.predicate}."
            )
        if self.max_columns != self.min_columns:
            common_columns = self.applied_common_columns(lhs, rhs)
            operation = dataclasses.replace(self, min_columns=common_columns, max_columns=common_columns)
        else:
            if not lhs.columns >= self.common_columns:
                raise ColumnError(
                    f"Missing columns {set(self.common_columns - lhs.columns)} "
                    f"for left-hand side of join between {lhs!r} and {rhs!r}."
                )
            if not rhs.columns >= self.common_columns:
                raise ColumnError(
                    f"Missing columns {set(self.common_columns - rhs.columns)} "
                    f"for right-hand side of join between {lhs!r} and {rhs!r}."
                )
            operation = self
        if lhs.is_join_identity:
            return IgnoreOne(True)
        if rhs.is_join_identity:
            return IgnoreOne(False)
        return operation

    def _finish_apply(self, lhs: Relation, rhs: Relation) -> Relation:
        # Docstring inherited.
        if lhs.is_join_identity:
            return rhs
        if rhs.is_join_identity:
            return lhs
        if lhs.engine != rhs.engine:
            raise EngineError(f"Mismatched join engines: {lhs.engine} != {rhs.engine}.")
        if not self.predicate.is_supported_by(lhs.engine):
            raise EngineError(f"Join predicate {self.predicate} does not support engine {lhs.engine}.")
        return super()._finish_apply(lhs, rhs)

    def applied_columns(self, lhs: Relation, rhs: Relation) -> Set[ColumnTag]:
        # Docstring inherited.
        return lhs.columns | rhs.columns

    def applied_min_rows(self, lhs: Relation, rhs: Relation) -> int:
        # Docstring inherited.
        return 0

    def applied_max_rows(self, lhs: Relation, rhs: Relation) -> int | None:
        # Docstring inherited.
        if lhs.max_rows == 0 or rhs.max_rows == 0:
            return 0
        if lhs.max_rows is None or rhs.max_rows is None:
            return None
        else:
            return lhs.max_rows * rhs.max_rows

    def applied_common_columns(self, lhs: Relation, rhs: Relation) -> frozenset[ColumnTag]:
        """Compute the actual common columns for a `Join` given its targets.

        Parameters
        ----------
        lhs : `Relation`
            One relation to join.
        rhs : `Relation`
            The other relation to join to ``lhs``.

        Returns
        -------
        common_columns : `~collections.abc.Set` [ `ColumnTag` ]
            Columns that are included in all of ``lhs.columns`` and
            ``rhs.columns`` and `max_columns`, checked to be a superset of
            `min_columns`.

        Raises
        ------
        ColumnError
            Raised if the result would not be a superset of `min_columns`.
        """
        # Docstring inherited.
        if self.max_columns != self.min_columns:
            common_columns = {tag for tag in lhs.columns & rhs.columns if tag.is_key}
            if self.max_columns is not None:
                common_columns &= self.max_columns
            if not (common_columns >= self.min_columns):
                raise ColumnError(
                    f"Common columns {common_columns} for join between {lhs} and {rhs} are not a superset "
                    f"of the minimum columns {self.min_columns}."
                )
            return frozenset(common_columns)
        else:
            return self.min_columns

    def partial(self, fix: Relation, is_lhs: bool = False) -> PartialJoin:
        """Return a `UnaryOperation` that represents this join with one operand
        already provided and held fixed.

        Parameters
        ----------
        fix : `Relation`
            Relation to include in the returned unary operation.
        is_lhs : `bool`, optional
            Whether ``fix`` should be considered the ``lhs`` or ``rhs`` side of
            the join (`Join` side is *usually* irrelevant, but `Engine`
            implementations are permitted to make additional guarantees about
            row order or duplicates based on them).

        Returns
        -------
        partial_join : `PartialJoin`
            Unary operation representing a join to a fixed relation.

        Raises
        ------
        ColumnError
            Raised if the given predicate requires columns not present in
            ``lhs`` or ``rhs``.
        RowOrderError
            Raised if ``lhs`` or ``rhs`` is unnecessarily ordered; see
            `Relation.expect_unordered`.

        Notes
        -----
        This method and the class it returns are called "partial" in the spirit
        of `functools.partial`: a callable formed by holding some arguments to
        another callable fixed.
        """
        if not (self.min_columns <= fix.columns):
            raise ColumnError(
                f"Missing columns {set(self.min_columns - fix.columns)} for partial join to {fix}."
            )
        return PartialJoin(self, fix, is_lhs)


@final
@dataclasses.dataclass(frozen=True)
class PartialJoin(UnaryOperation):
    """A `UnaryOperation` that represents this join with one operand already
    provided and held fixed.

    Notes
    -----
    This class and the `Join.partial` used to construct it are called "partial"
    in the spirit of `functools.partial`: a callable formed by holding some
    arguments to another callable fixed.

    `PartialJoin` instances never appear in relation trees; the `apply` method
    will return a `BinaryOperationRelation` with a `Join` operation instead of
    a `UnaryOperationRelation` with a `PartialJoin` (or one of the operands, if
    the other is a `join identity relation <Relation.is_join_identity>`).
    """

    binary: Join
    """The join operation (`Join`) to be applied.
    """

    fixed: Relation
    """The target relation already included in the operation (`Relation`).
    """

    fixed_is_lhs: bool
    """Whether `fixed` should be considered the ``lhs`` or ``rhs`` side of
    the join.

    `Join` side is *usually* irrelevant, but `Engine` implementations are
    permitted to make additional guarantees about row order or duplicates based
    on them.
    """

    @property
    def columns_required(self) -> Set[ColumnTag]:
        # Docstring inherited.
        result = set(self.binary.predicate.columns_required)
        result.difference_update(self.fixed.columns)
        result.update(self.binary.min_columns)
        return result

    @property
    def is_empty_invariant(self) -> bool:
        # Docstring inherited.
        return False

    @property
    def is_count_invariant(self) -> bool:
        # Docstring inherited.
        return False

    def __str__(self) -> str:
        return f"{self.binary!s}[{self.fixed!s}]"

    def _begin_apply(
        self, target: Relation, preferred_engine: Engine | None
    ) -> tuple[UnaryOperation, Engine]:
        # Docstring inherited.
        if self.binary.max_columns != self.binary.min_columns:
            common_columns = self.binary.applied_common_columns(self.fixed, target)
            replacement = dataclasses.replace(
                self,
                binary=dataclasses.replace(
                    self.binary, min_columns=common_columns, max_columns=common_columns
                ),
            )
            return replacement._begin_apply(target, preferred_engine)
        if preferred_engine is None:
            preferred_engine = self.fixed.engine
        if not self.columns_required <= target.columns:
            raise ColumnError(
                f"Join {self} to relation {target} needs columns "
                f"{set(self.columns_required) - target.columns}."
            )
        return super()._begin_apply(target, preferred_engine)

    def _finish_apply(self, target: Relation) -> Relation:
        # Docstring inherited.
        if self.fixed_is_lhs:
            return self.binary.apply(self.fixed, target)
        else:
            return self.binary.apply(target, self.fixed)

    def applied_columns(self, target: Relation) -> Set[ColumnTag]:
        # Docstring inherited.
        if self.fixed_is_lhs:
            return self.binary.applied_columns(self.fixed, target)
        else:
            return self.binary.applied_columns(target, self.fixed)

    def applied_min_rows(self, target: Relation) -> int:
        # Docstring inherited.
        if self.fixed_is_lhs:
            return self.binary.applied_min_rows(self.fixed, target)
        else:
            return self.binary.applied_min_rows(target, self.fixed)

    def applied_max_rows(self, target: Relation) -> int | None:
        # Docstring inherited.
        if self.fixed_is_lhs:
            return self.binary.applied_max_rows(self.fixed, target)
        else:
            return self.binary.applied_max_rows(target, self.fixed)

    def commute(self, current: UnaryOperationRelation) -> UnaryCommutator:
        # Docstring inherited.
        from ._deduplication import Deduplication
        from ._projection import Projection

        match current.operation:
            case Deduplication():
                # A Join only commutes past Deduplication if the fixed relation
                # has unique rows, which is not something we can check right
                # now.
                return UnaryCommutator(
                    first=None,
                    second=current.operation,
                    done=False,
                    messages=("join-deduplication commutation is not supported",),
                )
            case Projection():
                # In order for projection(join(target)) to be equivalent to
                # join(projection(target)), the new outer projection has to
                # include the columns added by the join.  Note that because we
                # require common_columns to be explicit at this point, the
                # projection cannot change them.
                return UnaryCommutator(
                    first=self,
                    second=Projection(frozenset(self.applied_columns(current))),
                )
            case _:
                if not self.columns_required <= current.target.columns:
                    return UnaryCommutator(
                        first=None,
                        second=current.operation,
                        done=False,
                        messages=(
                            f"{current.target} is missing columns "
                            f"{set(self.columns_required - current.target.columns)}",
                        ),
                    )
                if current.operation.is_count_dependent:
                    return UnaryCommutator(
                        first=None,
                        second=current.operation,
                        done=False,
                        messages=(f"{current.operation} is count-dependent",),
                    )
                return UnaryCommutator(first=self, second=current.operation)



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest

from lsst.daf.relation import (
    Calculation,
    ColumnError,
    ColumnExpression,
    EngineError,
    SortTerm,
    UnaryOperationRelation,
    iteration,
    tests,
)


class CalculationTestCase(tests.RelationTestCase):
    """Tests for the Calculation operation and relations based on it."""

    def setUp(self) -> None:
        self.a = tests.ColumnTag("a")
        self.b = tests.ColumnTag("b")
        self.expression = ColumnExpression.function(
            "__add__",
            ColumnExpression.reference(self.a),
            ColumnExpression.literal(10),
        )
        self.engine = iteration.Engine(name="preferred")
        self.leaf = self.engine.make_leaf(
            {self.a}, payload=iteration.RowSequence([{self.a: 0}, {self.a: 1}]), name="leaf"
        )

    def test_attributes(self) -> None:
        """Check that all UnaryOperation and Relation attributes have the
        expected values.
        """
        relation = self.leaf.with_calculated_column(self.b, self.expression)
        assert isinstance(relation, UnaryOperationRelation)
        self.assertEqual(relation.columns, {self.a, self.b})
        self.assertEqual(relation.engine, self.engine)
        self.assertEqual(relation.min_rows, self.leaf.min_rows)
        self.assertEqual(relation.max_rows, self.leaf.max_rows)
        operation = relation.operation
        assert isinstance(operation, Calculation)
        self.assertEqual(operation.expression, self.expression)
        self.assertEqual(operation.columns_required, {self.a})
        self.assertTrue(operation.is_empty_invariant)
        self.assertTrue(operation.is_count_invariant)
        self.assertFalse(operation.is_order_dependent)
        self.assertFalse(operation.is_count_dependent)

    def test_apply_failures(self) -> None:
        """Test failure modes of constructing and applying Calculations."""
        # Calculations must depend on existing columns.
        with self.assertRaises(ColumnError):
            self.leaf.with_calculated_column(self.b, ColumnExpression.literal(10))
        # Required columns must be present.
        with self.assertRaises(ColumnError):
            self.leaf.with_calculated_column(self.b, ColumnExpression.reference(tests.ColumnTag("c")))
        # New column must not already be present.
        with self.assertRaises(ColumnError):
            self.leaf.with_calculated_column(self.a, self.expression)

    def test_backtracking_apply(self) -> None:
        """Test apply logic that involves reordering operations in the existing
        tree to perform the new operation in a preferred engine.
        """
        new_engine = iteration.Engine(name="downstream")
        predicate = ColumnExpression.reference(self.b).lt(ColumnExpression.literal(20))
        # Apply a bunch of operations in a new engine that a Calculation should
        # commute with.
        target = (
            self.leaf.transferred_to(new_engine)
            .with_calculated_column(self.b, self.expression)
            .with_rows_satisfying(predicate)
            .without_duplicates()
            .with_only_columns({self.a})
            .sorted([SortTerm(ColumnExpression.reference(self.a))])
        )[:2]
        # Apply a new Calculation with backtracking and see that it appears
        # before the transfer to the new engine, with adjustments as needed
        # downstream (to the Projection and Chain, in this case).
        tag = tests.ColumnTag("tag")
        relation = target.with_calculated_column(
            tag, self.expression, preferred_engine=self.engine, require_preferred_engine=True
        )
        self.assert_relations_equal(
            relation,
            (
                self.leaf.with_calculated_column(tag, self.expression)
                .transferred_to(new_engine)
                .with_calculated_column(self.b, self.expression)
                .with_rows_satisfying(predicate)
                .without_duplicates()
                .with_only_columns({self.a, tag})
                .sorted([SortTerm(ColumnExpression.reference(self.a))])
            )[:2],
        )

    def test_no_backtracking(self) -> None:
        """Test apply logic that handles preferred engines without reordering
        operations in the existing tree.
        """
        new_engine = iteration.Engine(name="downstream")
        # Construct a relation tree we can't reorder when inserting a
        # Calculation, for various reasons.
        leaf = self.engine.make_leaf(
            {self.a},
            payload=iteration.RowSequence([{self.a: 100}, {self.a: 200}]),
            name="leaf",
        )
        # Can't insert after leaf because Materialization is locked.
        target = leaf.transferred_to(new_engine).materialized("lock")
        # Preferred engine is ignored if we can't backtrack and don't enable
        # anything else.
        self.assert_relations_equal(
            target.with_calculated_column(self.b, self.expression, preferred_engine=self.engine),
            target.with_calculated_column(self.b, self.expression),
        )
        # We can force this to be an error.
        with self.assertRaises(EngineError):
            target.with_calculated_column(
                self.b, self.expression, preferred_engine=self.engine, require_preferred_engine=True
            )
        # We can also automatically transfer (back) to the preferred engine.
        self.assert_relations_equal(
            target.with_calculated_column(
                self.b, self.expression, preferred_engine=self.engine, transfer=True
            ),
            target.transferred_to(self.engine).with_calculated_column(self.b, self.expression),
        )
        # Can't commute past an existing Calculation that provides a required
        # column.
        target = leaf.transferred_to(new_engine).with_calculated_column(self.b, self.expression)
        with self.assertRaises(EngineError):
            target.with_calculated_column(
                tests.ColumnTag("c"),
                ColumnExpression.reference(self.b),
                preferred_engine=self.engine,
                require_preferred_engine=True,
            )

    def test_iteration(self) -> None:
        """Test Calculation execution in the iteration engine."""
        relation = self.leaf.with_calculated_column(self.b, self.expression)
        self.assertEqual(
            list(self.engine.execute(relation)),
            [{self.a: 0, self.b: 10}, {self.a: 1, self.b: 11}],
        )

    def test_str(self) -> None:
        """Test str(Calculation) and
        str(UnaryOperationRelation[Calculation]).
        """
        relation = self.leaf.with_calculated_column(self.b, self.expression)
        self.assertEqual(str(relation), "+[b=__add__(a, 10)](leaf)")


if __name__ == "__main__":
    unittest.main()
